;; Copyright 2009, 2010, 2011, 2012 Kevin Ryde
;;
;; This file is part of accjournal.
;;
;; accjournal is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; accjournal is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with accjournal; see the file COPYING.  Failing that, go to
;; <http://www.gnu.org/licenses/>.


;; This is a few lines to convert accjournal format to ledger format
;; http://newartisans.com/software/ledger.html
;;
;; It's a bit rough, but if you started in accjournal and want to give
;; ledger a try then it can do most of the conversion.
;;
;; `accjournal-mung-account-name-list' can convert account names to
;; something more ledger-like.  The manglings shown are only examples, add
;; or delete as necessary to get the ledger chart of accounts coming out as
;; desired.
;;
;; Other things for the output can be manipulated by changing the code.
;; It's the sort of thing only run once or twice, so there's no need for a
;; lot of customizations.

;; Usage:
;;
;; To make a conversion, in this .el file
;;
;;     M-x eval-buffer
;; or
;;     (load-file "accjournal-to-ledger.el")
;;
;; then in an accjournal text buffer
;;
;;     M-x accjournal-to-ledger
;;
;; which makes a new buffer "*accjournal-to-ledger*" which is in ledger
;; format and can be saved to a desired file.


(require 'cl)
(require 'accjournal)

(defvar accjournal-mung-account-name-width 28)

(defvar accjournal-mung-account-name-list
  '(
    ;; "xInitial" or "zInitial" -> "Equity:Opening"
    ("\\`[xz]Initial\\'" . "Equity:Opening")

    ;; "BHP 1000(a)" -> "Assets:BHP 1000(a)"
    ;; shares with a name and number
    (" [0-9][0-9]+\\(([a-z]+)\\)?\\(\\'\\| - [A-Z]\\)"
     . (lambda (name) (concat "Assets:" name)))

    ;; "Savings 06/07" -> "Tax-06/07:Savings"
    ;; bank accounts recorded by tax year
    (" \\([0-9][0-9]/[0-9][0-9]\\)\\'"
     . (lambda (name) (concat "Tax-" (match-string 1 name)
                              ":" (substring name 0 (match-beginning 0)))))

    ;; "06/07 cap gains" -> "Tax-06/07:cap gains"
    ;; tax year 06/07 etc calculations
    ("\\`\\([0-9][0-9]/[0-9][0-9]\\) "
     . (lambda (name) (concat "Tax-" (match-string 1 name)
                              ":" (substring name (match-end 0)))))

    ;; "Settle 28sep10" -> "Temp:Settle 28sep10"
    ;; with a date is a temporary account
    (" [0-9][0-9][a-zA-Z][a-zA-Z][a-zA-Z][0-9][0-9]\\'"
     . (lambda (name) (concat "Temp:" name))))

  "List of manglings for accjournal to ledger account names.
Each element is a pair

    (REGEXP . STRING)
    (REGEXP . FUNC)

An account name matching REGEXP is changed to either a given
fixed STRING or to the value from a FUNC called with the account
name string

    (FUNC name)

The `match-data' is from matching REGEXP on name, so the \\(..\\)
groups in REGEXP are `(match-string 1 name)' etc.

If FUNC returns nil it's taken that it didn't want to modify the
given name.

All the FUNCs in the samples are anon (lambda (name) ...) forms.
A symbol for a named function is fine too, but an anonymous
function for a couple of lines of mangling should be enough.")

(defun accjournal-mung-account-name (name)
  "Apply `accjournal-mung-account-name-list' to NAME (a string).
The return is a new string, or perhaps NAME unchanged."
  (loop for (regexp . func) in accjournal-mung-account-name-list
        do (if (string-match regexp name)
               (setq name (if (stringp func)
                             func
                           (or (funcall func name)
                               name))))
        finally return name))

(defun accjournal-date-to-iso (str)
  "STR per `accjournal-date-to-N' returned as iso string \"2010-09-29\""
  (setq str (number-to-string (accjournal-date-to-N str)))
  (concat (substring str 0 4) "-" (substring str 4 6) "-" (substring str 6)))

(defun accjournal-to-ledger ()
  "Convert accjournal lines in the current buffer to ledger format.
Ledger format is in put in a new buffer \"*accjournal-to-ledger*\"."

  (interactive)
  (let ((output-buffer (get-buffer-create "*accjournal-to-ledger*")))
    (with-current-buffer output-buffer
      (erase-buffer))

    (goto-char (point-min))
    (let ((to-format (format "  %%-%ds $%%s\n"
                             accjournal-mung-account-name-width)))
      (progn ;; accjournal-push-mark-on-error
        (goto-char (point-min))
        (while (not (eobp))
          (let (out)
            (cond
             ((looking-at comment-start-skip)
              (setq out (concat ";; " (buffer-substring (match-end 0)
                                                        (line-end-position)))))

             ((looking-at "include[ \t]\\(\"?\\)\\(.*\\)\\1")
              (let ((filename (match-string 2)))
                (setq filename (replace-regexp-in-string
                                "\\.txt\\'" ".ledger" filename t t))
                (setq out (concat "!include " filename))))

             ((looking-at accjournal-line-regexp)
              (let ((date           (match-string 1))

                    (amount-pos     (match-beginning  2))
                    (amount         (match-string     2))
                    (amount-dollars (match-string     3))
                    (amount-decimal (match-string     4))
                    (amount-cents   (or (match-string 5) ""))

                    (from-pos       (or (match-beginning 6)
                                        (match-beginning 7)))
                    (from-open      (match-beginning 6))
                    (from-name      (match-string    7))
                    (from-close     (match-beginning 8))

                    (to-pos         (or (match-beginning 9)
                                        (match-beginning 10)))
                    (to-open        (match-beginning 9))
                    (to-name        (match-string 10))
                    (to-close       (match-beginning 11))

                    (note           (match-string 12)))

                (setq date (accjournal-date-to-iso date))

                (setq from-name (accjournal-mung-account-name from-name))
                (setq to-name   (accjournal-mung-account-name to-name))

                ;; nothing in ledger to say this is the start of an account
                ;;    (if (equal "!" from-open)
                ;;        (accjournal-open from-name))
                ;;    (if (equal "!" to-open)
                ;;        (accjournal-open to-name))

                (setq out (concat (format "%s  %s\n" date note)
                                  (format to-format to-name amount)
                                  (format "  %s" from-name)))

                ;; nothing in ledger to say this is the end of an account
                ;;   (if (equal "@" from-close)
                ;;       (accjournal-close from))
                ;;   (if (equal "@" to-close)
                ;;       (accjournal-close to-name))))

                ))

             (t
              (setq out (buffer-substring (line-beginning-position)
                                          (line-end-position)))))
            (with-current-buffer output-buffer
              (insert out "\n")))

          (forward-line))))

    (switch-to-buffer output-buffer)

    (goto-char (point-min))
    (while (search-forward "-*- mode: accjournal -*-" nil t)
      (replace-match "-*- mode: ledger -*-" t t)) ;; fixedcase, literal

    (goto-char (point-min))
    (ledger-mode)))
